#include <Wire.h>
#include <Adafruit_ADS1015.h>
#include <PubSubClient.h>
#include <ESP8266WiFi.h>
#include <OneWire.h>
#include <DallasTemperature.h>
#include <FastLED.h>
#include <ESP8266HTTPClient.h>
#include <ArduinoJson.h>
#include <time.h>

#define MQTT_SERVER "192.168.1.3"
#define SSID "GCHQ Surveillance Van"
#define PASSWORD "cocklol."
#define LIGHT_TOPIC "/home/switches/kitchen/fishtank"
#define LIGHT_PIN D2
#define LED_COUNT 36
#define PH_TOPIC "/home/sensors/fishtank/ph"
#define TEMPERATURE_TOPIC "/home/sensors/fishtank/temperature"
#define TEMPERATURE_PIN D5

#define PH_7_VOLTAGE 2.5
#define PH_4_VOLTAGE 3.04

#define VOLTAGE_OFFSET 0.03

#define SOLAR_EVENT_URL "http://api.sunrise-sunset.org/json?lat=20.548103&lng=96.916835&formatted=0"
#define TIMEZONE_OFFSET 23400 // 6.5 hours in seconds
//#define TIME_SHIFT 5400 // 1.5 hours
#define NTP_POOL "uk.pool.ntp.org"

void mqttCallback(char *topic, byte *payload, unsigned int length);

Adafruit_ADS1115 ads;
OneWire oneWire(TEMPERATURE_PIN);
DallasTemperature DS18B20(&oneWire);
WiFiClient wifiClient;
PubSubClient client = PubSubClient(MQTT_SERVER, 1883, mqttCallback, wifiClient);
CRGB leds[LED_COUNT];
bool lightsOn = true;

static const uint8_t sunriseLength = 30;
static const uint8_t heatIndexMax = 241; // After 241 the pallet seems to go dim
static const float lightChangeInterval = ((float)(sunriseLength * 60) / heatIndexMax) * 1000;
static uint8_t heatIndex = 0;
int _lightsOnStartTime, _lightsOnEndTime, _lightsOffStartTime, _lightsOffEndTime;

float pHStep = (PH_7_VOLTAGE - PH_4_VOLTAGE) / 3;

void waitForWiFi() {
	if (WiFi.status() == WL_CONNECTED)
		return;
	Serial.println("Reconnecting WiFi...");
	int waitCount=0;
	while (WiFi.status() != WL_CONNECTED) {
		delay(100);
		waitCount++;
		if (waitCount>600)
			ESP.restart();
	}
	Serial.println("Connected WiFi!");
}

void waitForMQTT() {
	if (client.connected())
		return;
	Serial.println("Reconnecting MQTT...");
	while (!client.connected())
		if (!client.connect(PH_TOPIC))
			delay(50);
	client.subscribe(LIGHT_TOPIC);
	Serial.println(LIGHT_TOPIC);
	Serial.println("Connected MQTT!");
}

void reconnect() {
	waitForWiFi();
	waitForMQTT();
}

String getSolarEventJson(){
	waitForWiFi();

	HTTPClient http;
	http.begin(SOLAR_EVENT_URL);
	int httpCode = http.GET();
	String result = "";
	if (httpCode > 0)
		result = http.getString();
	http.end();

	return result;
}

int getEpoch(const char *str) {
	struct tm time;
	int year, month;
	sscanf(str, "%d-%d-%dT%d:%d:%d+00:00", &year, &month, &time.tm_mday, &time.tm_hour, &time.tm_min, &time.tm_sec);
	time.tm_year = year - 1900;
	time.tm_mon = month - 1;
	return mktime(&time);
}

int adjustForTimezone(int epoch){
	return epoch + TIMEZONE_OFFSET;
}

void parseEvents(String json){
	StaticJsonBuffer<1024> jsonBuffer;
	JsonObject &root = jsonBuffer.parseObject(json);

	if (!root.success()) {
		Serial.println("Could not parse json");
		return;
	}

	const char *astronomicalTwilightBegin = root["results"]["astronomical_twilight_begin"];
	const char *sunrise = root["results"]["sunrise"];
	const char *sunset = root["results"]["sunset"];
	const char *astronomicalTwilightEnd = root["results"]["astronomical_twilight_end"];

	_lightsOnStartTime = adjustForTimezone(getEpoch(astronomicalTwilightBegin));
	_lightsOnEndTime = adjustForTimezone(getEpoch(sunrise));
	_lightsOffStartTime = adjustForTimezone(getEpoch(sunset));
	_lightsOffEndTime = adjustForTimezone(getEpoch(astronomicalTwilightEnd));
	Serial.println(_lightsOffEndTime);
}

void setup() {
	Serial.begin(115200);

	ads.setGain(GAIN_TWOTHIRDS);
	ads.begin();

	pinMode(TEMPERATURE_PIN, INPUT_PULLUP);
	DS18B20.begin();

	WiFi.begin(SSID, PASSWORD);
	reconnect();

	delay(2000);
	configTime(0, 0, NTP_POOL);

	String json = getSolarEventJson();
	parseEvents(json);

	FastLED.addLeds<WS2811, LIGHT_PIN, GRB>(leds, LED_COUNT).setCorrection(TypicalSMD5050).setTemperature(Tungsten40W);

	Serial.print("Light interval: ");
	Serial.println(lightChangeInterval);
}

float readpH() {
	Serial.println("Reading pH...");
	int sum = 0;
	const int readCount = 10;

	for (int i = 0; i < readCount; i++)
	{
		sum += ads.readADC_SingleEnded(0);
		delay(10);
	}

	float averageRead = float(sum) / readCount;
	Serial.print("Average value: ");
	Serial.println(averageRead);

	float voltage = 6.144 / 32768.0 * averageRead;
	voltage -= VOLTAGE_OFFSET;
	Serial.println(voltage);
	float pH = 7 - ((PH_7_VOLTAGE - voltage) / pHStep);
	return pH;
}

void publishpH() {
	float pH = readpH();
	String pHString(pH, 2);
	Serial.println(pHString);
	client.publish(PH_TOPIC, pHString.c_str(), true);
}

void publishTemperature() {
	DS18B20.requestTemperatures();
	float temperature = DS18B20.getTempCByIndex(0);
	String temperatureString(temperature, 2);
	Serial.println(temperatureString);
	client.publish(TEMPERATURE_TOPIC, temperatureString.c_str(), true);
}

void doLighting() {
	EVERY_N_MILLISECONDS(lightChangeInterval) {
		CRGB colour = ColorFromPalette(HeatColors_p, heatIndex);

		fill_solid(leds, LED_COUNT, colour);
		if (lightsOn){
			if(heatIndex < heatIndexMax)
				heatIndex++;
		} else {
			if (heatIndex > 0)
				heatIndex--;
		}

		FastLED.show();
	}
}

void publishReadings(){
	EVERY_N_SECONDS(60) {
		publishpH();
		publishTemperature();
	}
}

void loop() {
	reconnect();
	client.loop();

	doLighting();

	//publishReadings();

	delay(10);
}

void mqttCallback(char *topic, byte *payload, unsigned int length) {
	Serial.print("Got payload: ");
	for (int i = 0; i < length; i++)
	{
		Serial.print((char)payload[i]);
	}
	Serial.println();

	lightsOn = payload[0] == '1';
}